@@ -48,6 +48,8 @@
 #include <usterr-signal-safe.h>
 #include <helper.h>
 #include "lttng-tracer.h"
+#include "../libringbuffer/shm.h"
+#include "../libringbuffer/frontend_types.h"
 
 #define OBJ_NAME_LEN	16
 
@@ -69,6 +71,7 @@ struct lttng_ust_obj {
 			void *private_data;
 			const struct lttng_ust_objd_ops *ops;
 			int f_count;
+			int owner_ref;	/* has ref from owner */
 			void *owner;
 			char name[OBJ_NAME_LEN];
 		} s;
@@ -124,6 +127,7 @@ end:
 	obj->u.s.ops = ops;
 	obj->u.s.f_count = 2;	/* count == 1 : object is allocated */
 				/* count == 2 : allocated + hold ref */
+	obj->u.s.owner_ref = 1;	/* One owner reference */
 	obj->u.s.owner = owner;
 	strncpy(obj->u.s.name, name, OBJ_NAME_LEN);
 	obj->u.s.name[OBJ_NAME_LEN - 1] = '\0';
@@ -184,7 +188,7 @@ void objd_ref(int id)
 	obj->u.s.f_count++;
 }
 
-int lttng_ust_objd_unref(int id)
+int lttng_ust_objd_unref(int id, int is_owner)
 {
 	struct lttng_ust_obj *obj = _objd_get(id);
 
@@ -194,6 +198,13 @@ int lttng_ust_objd_unref(int id)
 		ERR("Reference counting error\n");
 		return -EINVAL;
 	}
+	if (is_owner) {
+		if (!obj->u.s.owner_ref) {
+			ERR("Error decrementing owner reference");
+			return -EINVAL;
+		}
+		obj->u.s.owner_ref--;
+	}
 	if ((--obj->u.s.f_count) == 1) {
 		const struct lttng_ust_objd_ops *ops = objd_ops(id);
 		
@@ -209,8 +220,16 @@ void objd_table_destroy(void)
 {
 	int i;
 
-	for (i = 0; i < objd_table.allocated_len; i++)
-		(void) lttng_ust_objd_unref(i);
+	for (i = 0; i < objd_table.allocated_len; i++) {
+		struct lttng_ust_obj *obj;
+
+		obj = _objd_get(i);
+		if (!obj)
+			continue;
+		if (!obj->u.s.owner_ref)
+			continue;	/* only unref owner ref. */
+		(void) lttng_ust_objd_unref(i, 1);
+	}
 	free(objd_table.array);
 	objd_table.array = NULL;
 	objd_table.len = 0;
@@ -218,6 +237,15 @@ void objd_table_destroy(void)
 	objd_table.freelist_head = -1;
 }
 
+const char *lttng_ust_obj_get_name(int id)
+{
+	struct lttng_ust_obj *obj = _objd_get(id);
+
+	if (!obj)
+		return NULL;
+	return obj->u.s.name;
+}
+
 void lttng_ust_objd_table_owner_cleanup(void *owner)
 {
 	int i;
@@ -230,8 +258,10 @@ void lttng_ust_objd_table_owner_cleanup(void *owner)
 			continue;
 		if (!obj->u.s.owner)
 			continue;	/* skip root handles */
+		if (!obj->u.s.owner_ref)
+			continue;	/* only unref owner ref. */
 		if (obj->u.s.owner == owner)
-			(void) lttng_ust_objd_unref(i);
+			(void) lttng_ust_objd_unref(i, 1);
 	}
 }
 
@@ -243,17 +273,10 @@ void lttng_ust_objd_table_owner_cleanup(void *owner)
 static const struct lttng_ust_objd_ops lttng_ops;
 static const struct lttng_ust_objd_ops lttng_session_ops;
 static const struct lttng_ust_objd_ops lttng_channel_ops;
-static const struct lttng_ust_objd_ops lttng_metadata_ops;
 static const struct lttng_ust_objd_ops lttng_enabler_ops;
-static const struct lttng_ust_objd_ops lib_ring_buffer_objd_ops;
 static const struct lttng_ust_objd_ops lttng_tracepoint_list_ops;
 static const struct lttng_ust_objd_ops lttng_tracepoint_field_list_ops;
 
-enum channel_type {
-	PER_CPU_CHANNEL,
-	METADATA_CHANNEL,
-};
-
 int lttng_abi_create_root_handle(void)
 {
 	int root_handle;
@@ -263,6 +286,18 @@ int lttng_abi_create_root_handle(void)
 	return root_handle;
 }
 
+static
+int lttng_is_channel_ready(struct lttng_channel *lttng_chan)
+{
+	struct channel *chan;
+	unsigned int nr_streams, exp_streams;
+
+	chan = lttng_chan->chan;
+	nr_streams = channel_handle_get_nr_streams(lttng_chan->handle);
+	exp_streams = chan->nr_streams;
+	return nr_streams == exp_streams;
+}
+
 static
 int lttng_abi_create_session(void *owner)
 {
@@ -278,6 +313,7 @@ int lttng_abi_create_session(void *owner)
 		goto objd_error;
 	}
 	session->objd = session_objd;
+	session->owner = owner;
 	return session_objd;
 
 objd_error:
@@ -289,9 +325,9 @@ static
 long lttng_abi_tracer_version(int objd,
 	struct lttng_ust_tracer_version *v)
 {
-	v->major = LTTNG_UST_INTERNAL_MAJOR_VERSION;
-	v->minor = LTTNG_UST_INTERNAL_MINOR_VERSION;
-	v->patchlevel = LTTNG_UST_INTERNAL_PATCHLEVEL_VERSION;
+	v->major = LTTNG_UST_MAJOR_VERSION;
+	v->minor = LTTNG_UST_MINOR_VERSION;
+	v->patchlevel = LTTNG_UST_PATCHLEVEL_VERSION;
 	return 0;
 }
 
@@ -352,119 +388,152 @@ static const struct lttng_ust_objd_ops lttng_ops = {
 	.cmd = lttng_cmd,
 };
 
-/*
- * We tolerate no failure in this function (if one happens, we print a dmesg
- * error, but cannot return any error, because the channel information is
- * invariant.
- */
-static
-void lttng_metadata_create_events(int channel_objd)
-{
-	struct lttng_channel *chan = objd_private(channel_objd);
-	struct lttng_enabler *enabler;
-	static struct lttng_ust_event metadata_params = {
-		.instrumentation = LTTNG_UST_TRACEPOINT,
-		.name = "lttng_ust:metadata",
-		.loglevel_type = LTTNG_UST_LOGLEVEL_ALL,
-		.loglevel = TRACE_DEFAULT,
-	};
-
-	/*
-	 * We tolerate no failure path after event creation. It will stay
-	 * invariant for the rest of the session.
-	 */
-	enabler = lttng_enabler_create(LTTNG_ENABLER_EVENT,
-				&metadata_params, chan);
-	if (!enabler) {
-		goto create_error;
-	}
-	return;
-
-create_error:
-	WARN_ON(1);
-	return;		/* not allowed to return error */
-}
-
-int lttng_abi_create_channel(int session_objd,
-			     struct lttng_ust_channel *chan_param,
-			     enum channel_type channel_type,
-			     union ust_args *uargs,
-			     void *owner)
+int lttng_abi_map_channel(int session_objd,
+		struct lttng_ust_channel *ust_chan,
+		union ust_args *uargs,
+		void *owner)
 {
 	struct lttng_session *session = objd_private(session_objd);
-	const struct lttng_ust_objd_ops *ops;
 	const char *transport_name;
-	struct lttng_channel *chan;
+	const struct lttng_transport *transport;
+	const char *chan_name;
 	int chan_objd;
-	int ret = 0;
-	struct lttng_channel chan_priv_init;
+	struct lttng_ust_shm_handle *channel_handle;
+	struct lttng_channel *lttng_chan;
+	struct channel *chan;
+	struct lttng_ust_lib_ring_buffer_config *config;
+	void *chan_data;
+	int wakeup_fd;
+	uint64_t len;
+	int ret;
+	enum lttng_ust_chan_type type;
 
-	switch (channel_type) {
-	case PER_CPU_CHANNEL:
-		if (chan_param->output == LTTNG_UST_MMAP) {
-			transport_name = chan_param->overwrite ?
-				"relay-overwrite-mmap" : "relay-discard-mmap";
-		} else {
-			return -EINVAL;
-		}
-		ops = &lttng_channel_ops;
+	chan_data = uargs->channel.chan_data;
+	wakeup_fd = uargs->channel.wakeup_fd;
+	len = ust_chan->len;
+	type = ust_chan->type;
+
+	switch (type) {
+	case LTTNG_UST_CHAN_PER_CPU:
 		break;
-	case METADATA_CHANNEL:
-		if (chan_param->output == LTTNG_UST_MMAP)
-			transport_name = "relay-metadata-mmap";
-		else
-			return -EINVAL;
-		ops = &lttng_metadata_ops;
+	default:
+		ret = -EINVAL;
+		goto invalid;
+	}
+
+	if (session->been_active) {
+		ret = -EBUSY;
+		goto active;	/* Refuse to add channel to active session */
+	}
+
+	channel_handle = channel_handle_create(chan_data, len, wakeup_fd);
+	if (!channel_handle) {
+		ret = -EINVAL;
+		goto handle_error;
+	}
+
+	chan = shmp(channel_handle, channel_handle->chan);
+	assert(chan);
+	chan->handle = channel_handle;
+	config = &chan->backend.config;
+	lttng_chan = channel_get_private(chan);
+	if (!lttng_chan) {
+		ret = -EINVAL;
+		goto alloc_error;
+	}
+
+	/* Lookup transport name */
+	switch (type) {
+	case LTTNG_UST_CHAN_PER_CPU:
+		if (config->output == RING_BUFFER_MMAP) {
+			if (config->mode == RING_BUFFER_OVERWRITE) {
+				if (config->wakeup == RING_BUFFER_WAKEUP_BY_WRITER) {
+					transport_name = "relay-overwrite-mmap";
+				} else {
+					transport_name = "relay-overwrite-rt-mmap";
+				}
+			} else {
+				if (config->wakeup == RING_BUFFER_WAKEUP_BY_WRITER) {
+					transport_name = "relay-discard-mmap";
+				} else {
+					transport_name = "relay-discard-rt-mmap";
+				}
+			}
+		} else {
+			ret = -EINVAL;
+			goto notransport;
+		}
+		chan_name = "channel";
 		break;
 	default:
 		transport_name = "<unknown>";
-		return -EINVAL;
+		chan_name = "<unknown>";
+		ret = -EINVAL;
+		goto notransport;
 	}
-	chan_objd = objd_alloc(NULL, ops, owner, "channel");
+	transport = lttng_transport_find(transport_name);
+	if (!transport) {
+		DBG("LTTng transport %s not found\n",
+		       transport_name);
+		ret = -EINVAL;
+		goto notransport;
+	}
+
+	chan_objd = objd_alloc(NULL, &lttng_channel_ops, owner, chan_name);
 	if (chan_objd < 0) {
 		ret = chan_objd;
 		goto objd_error;
 	}
-	memset(&chan_priv_init, 0, sizeof(chan_priv_init));
-	/* Copy of session UUID for consumer (availability through shm) */
-	memcpy(chan_priv_init.uuid, session->uuid, sizeof(session->uuid));
-	
+
+	/* Initialize our lttng chan */
+	lttng_chan->chan = chan;
+	lttng_chan->tstate = 1;
+	lttng_chan->enabled = 1;
+	lttng_chan->ctx = NULL;
+	lttng_chan->session = session;
+	lttng_chan->ops = &transport->ops;
+	memcpy(&lttng_chan->chan->backend.config,
+		transport->client_config,
+		sizeof(lttng_chan->chan->backend.config));
+	cds_list_add(&lttng_chan->node, &session->chan_head);
+	lttng_chan->header_type = 0;
+	lttng_chan->handle = channel_handle;
+	lttng_chan->type = type;
+
 	/*
 	 * We tolerate no failure path after channel creation. It will stay
 	 * invariant for the rest of the session.
 	 */
-	chan = lttng_channel_create(session, transport_name, NULL,
-				  chan_param->subbuf_size,
-				  chan_param->num_subbuf,
-				  chan_param->switch_timer_interval,
-				  chan_param->read_timer_interval,
-				  &uargs->channel.shm_fd,
-				  &uargs->channel.wait_fd,
-				  &uargs->channel.memory_map_size,
-				  &chan_priv_init);
-	if (!chan) {
-		ret = -EINVAL;
-		goto chan_error;
-	}
-	objd_set_private(chan_objd, chan);
-	chan->objd = chan_objd;
-	if (channel_type == METADATA_CHANNEL) {
-		session->metadata = chan;
-		lttng_metadata_create_events(chan_objd);
-	}
+	objd_set_private(chan_objd, lttng_chan);
+	lttng_chan->objd = chan_objd;
 	/* The channel created holds a reference on the session */
 	objd_ref(session_objd);
-
 	return chan_objd;
 
-chan_error:
-	{
-		int err;
-
-		err = lttng_ust_objd_unref(chan_objd);
-		assert(!err);
-	}
+	/* error path after channel was created */
 objd_error:
+notransport:
+	free(lttng_chan);
+alloc_error:
+	channel_destroy(chan, channel_handle, 0);
+	return ret;
+
+	/*
+	 * error path before channel creation (owning chan_data and
+	 * wakeup_fd).
+	 */
+handle_error:
+active:
+invalid:
+	{
+		int close_ret;
+
+		close_ret = close(wakeup_fd);
+		if (close_ret) {
+			PERROR("close");
+		}
+	}
+	free(chan_data);
 	return ret;
 }
 
@@ -484,8 +553,6 @@ objd_error:
  *		Enables tracing for a session (weak enable)
  *	LTTNG_UST_DISABLE
  *		Disables tracing for a session (strong disable)
- *	LTTNG_UST_METADATA
- *		Returns a LTTng metadata object descriptor
  *
  * The returned channel will be deleted when its file descriptor is closed.
  */
@@ -497,19 +564,15 @@ long lttng_session_cmd(int objd, unsigned int cmd, unsigned long arg,
 
 	switch (cmd) {
 	case LTTNG_UST_CHANNEL:
-		return lttng_abi_create_channel(objd,
+		return lttng_abi_map_channel(objd,
 				(struct lttng_ust_channel *) arg,
-				PER_CPU_CHANNEL, uargs, owner);
+				uargs, owner);
 	case LTTNG_UST_SESSION_START:
 	case LTTNG_UST_ENABLE:
 		return lttng_session_enable(session);
 	case LTTNG_UST_SESSION_STOP:
 	case LTTNG_UST_DISABLE:
 		return lttng_session_disable(session);
-	case LTTNG_UST_METADATA:
-		return lttng_abi_create_channel(objd,
-				(struct lttng_ust_channel *) arg,
-				METADATA_CHANNEL, uargs, owner);
 	default:
 		return -EINVAL;
 	}
@@ -553,12 +616,9 @@ long lttng_tracepoint_list_cmd(int objd, unsigned int cmd, unsigned long arg,
 	switch (cmd) {
 	case LTTNG_UST_TRACEPOINT_LIST_GET:
 	{
-	retry:
 		iter = lttng_ust_tracepoint_list_get_iter_next(list);
 		if (!iter)
 			return -LTTNG_UST_ERR_NOENT;
-		if (!strcmp(iter->name, "lttng_ust:metadata"))
-			goto retry;
 		memcpy(tp, iter, sizeof(*tp));
 		return 0;
 	}
@@ -598,7 +658,7 @@ alloc_error:
 	{
 		int err;
 
-		err = lttng_ust_objd_unref(list_objd);
+		err = lttng_ust_objd_unref(list_objd, 1);
 		assert(!err);
 	}
 objd_error:
@@ -635,12 +695,9 @@ long lttng_tracepoint_field_list_cmd(int objd, unsigned int cmd,
 	switch (cmd) {
 	case LTTNG_UST_TRACEPOINT_FIELD_LIST_GET:
 	{
-	retry:
 		iter = lttng_ust_field_list_get_iter_next(list);
 		if (!iter)
 			return -LTTNG_UST_ERR_NOENT;
-		if (!strcmp(iter->event_name, "lttng_ust:metadata"))
-			goto retry;
 		memcpy(tp, iter, sizeof(*tp));
 		return 0;
 	}
@@ -681,7 +738,7 @@ alloc_error:
 	{
 		int err;
 
-		err = lttng_ust_objd_unref(list_objd);
+		err = lttng_ust_objd_unref(list_objd, 1);
 		assert(!err);
 	}
 objd_error:
@@ -707,47 +764,22 @@ static const struct lttng_ust_objd_ops lttng_tracepoint_field_list_ops = {
 	.cmd = lttng_tracepoint_field_list_cmd,
 };
 
-struct stream_priv_data {
-	struct lttng_ust_lib_ring_buffer *buf;
-	struct lttng_channel *lttng_chan;
-};
-
 static
-int lttng_abi_open_stream(int channel_objd, struct lttng_ust_stream *info,
+int lttng_abi_map_stream(int channel_objd, struct lttng_ust_stream *info,
 		union ust_args *uargs, void *owner)
 {
 	struct lttng_channel *channel = objd_private(channel_objd);
-	struct lttng_ust_lib_ring_buffer *buf;
-	struct stream_priv_data *priv;
-	int stream_objd, ret;
+	int ret;
 
-	buf = channel->ops->buffer_read_open(channel->chan, channel->handle,
-			&uargs->stream.shm_fd,
-			&uargs->stream.wait_fd,
-			&uargs->stream.memory_map_size);
-	if (!buf)
-		return -ENOENT;
+	ret = channel_handle_add_stream(channel->handle,
+		uargs->stream.shm_fd, uargs->stream.wakeup_fd,
+		info->stream_nr, info->len);
+	if (ret)
+		goto error_add_stream;
 
-	priv = zmalloc(sizeof(*priv));
-	if (!priv) {
-		ret = -ENOMEM;
-		goto alloc_error;
-	}
-	priv->buf = buf;
-	priv->lttng_chan = channel;
-	stream_objd = objd_alloc(priv, &lib_ring_buffer_objd_ops, owner, "open_stream");
-	if (stream_objd < 0) {
-		ret = stream_objd;
-		goto objd_error;
-	}
-	/* Hold a reference on the channel object descriptor */
-	objd_ref(channel_objd);
-	return stream_objd;
+	return 0;
 
-objd_error:
-	free(priv);
-alloc_error:
-	channel->ops->buffer_read_close(buf, channel->handle);
+error_add_stream:
 	return ret;
 }
 
@@ -785,7 +817,7 @@ event_error:
 	{
 		int err;
 
-		err = lttng_ust_objd_unref(event_objd);
+		err = lttng_ust_objd_unref(event_objd, 1);
 		assert(!err);
 	}
 objd_error:
@@ -822,6 +854,14 @@ long lttng_channel_cmd(int objd, unsigned int cmd, unsigned long arg,
 {
 	struct lttng_channel *channel = objd_private(objd);
 
+	if (cmd != LTTNG_UST_STREAM) {
+		/*
+		 * Check if channel received all streams.
+		 */
+		if (!lttng_is_channel_ready(channel))
+			return -EPERM;
+	}
+
 	switch (cmd) {
 	case LTTNG_UST_STREAM:
 	{
@@ -829,7 +869,7 @@ long lttng_channel_cmd(int objd, unsigned int cmd, unsigned long arg,
 
 		stream = (struct lttng_ust_stream *) arg;
 		/* stream used as output */
-		return lttng_abi_open_stream(objd, stream, uargs, owner);
+		return lttng_abi_map_stream(objd, stream, uargs, owner);
 	}
 	case LTTNG_UST_EVENT:
 	{
@@ -859,50 +899,13 @@ long lttng_channel_cmd(int objd, unsigned int cmd, unsigned long arg,
 	}
 }
 
-/**
- *	lttng_metadata_cmd - lttng control through object descriptors
- *
- *	@objd: the object descriptor
- *	@cmd: the command
- *	@arg: command arg
- *	@uargs: UST arguments (internal)
- *	@owner: objd owner
- *
- *	This object descriptor implements lttng commands:
- *      LTTNG_UST_STREAM
- *              Returns an event stream file descriptor or failure.
- *
- * Channel and event file descriptors also hold a reference on the session.
- */
-static
-long lttng_metadata_cmd(int objd, unsigned int cmd, unsigned long arg,
-	union ust_args *uargs, void *owner)
-{
-	struct lttng_channel *channel = objd_private(objd);
-
-	switch (cmd) {
-	case LTTNG_UST_STREAM:
-	{
-		struct lttng_ust_stream *stream;
-
-		stream = (struct lttng_ust_stream *) arg;
-		/* stream used as output */
-		return lttng_abi_open_stream(objd, stream, uargs, owner);
-	}
-	case LTTNG_UST_FLUSH_BUFFER:
-		return channel->ops->flush_buffer(channel->chan, channel->handle);
-	default:
-		return -EINVAL;
-	}
-}
-
 static
 int lttng_channel_release(int objd)
 {
 	struct lttng_channel *channel = objd_private(objd);
 
 	if (channel)
-		return lttng_ust_objd_unref(channel->session->objd);
+		return lttng_ust_objd_unref(channel->session->objd, 0);
 	return 0;
 }
 
@@ -911,70 +914,6 @@ static const struct lttng_ust_objd_ops lttng_channel_ops = {
 	.cmd = lttng_channel_cmd,
 };
 
-static const struct lttng_ust_objd_ops lttng_metadata_ops = {
-	.release = lttng_channel_release,
-	.cmd = lttng_metadata_cmd,
-};
-
-/**
- *	lttng_rb_cmd - lttng ring buffer control through object descriptors
- *
- *	@objd: the object descriptor
- *	@cmd: the command
- *	@arg: command arg
- *	@uargs: UST arguments (internal)
- *	@owner: objd owner
- *
- *	This object descriptor implements lttng commands:
- *		(None for now. Access is done directly though shm.)
- */
-static
-long lttng_rb_cmd(int objd, unsigned int cmd, unsigned long arg,
-	union ust_args *uargs, void *owner)
-{
-	switch (cmd) {
-	default:
-		return -EINVAL;
-	}
-}
-
-static
-int lttng_rb_release(int objd)
-{
-	struct stream_priv_data *priv = objd_private(objd);
-	struct lttng_ust_lib_ring_buffer *buf;
-	struct lttng_channel *channel;
-
-	if (priv) {
-		buf = priv->buf;
-		channel = priv->lttng_chan;
-		free(priv);
-		/*
-		 * If we are at ABI exit, we don't want to close the
-		 * buffer opened for read: it is being shared between
-		 * the parent and child (right after fork), and we don't
-		 * want the child to close it for the parent. For a real
-		 * exit, we don't care about marking it as closed, as
-		 * the consumer daemon (if there is one) will do fine
-		 * even if we don't mark it as "closed" for reading on
-		 * our side.
-		 * We only mark it as closed if it is being explicitely
-		 * released by the session daemon with an explicit
-		 * release command.
-		 */
-		if (!lttng_ust_abi_close_in_progress)
-			channel->ops->buffer_read_close(buf, channel->handle);
-
-		return lttng_ust_objd_unref(channel->objd);
-	}
-	return 0;
-}
-
-static const struct lttng_ust_objd_ops lib_ring_buffer_objd_ops = {
-	.release = lttng_rb_release,
-	.cmd = lttng_rb_cmd,
-};
-
 /**
  *	lttng_enabler_cmd - lttng control through object descriptors
  *
@@ -1030,7 +969,7 @@ int lttng_enabler_release(int objd)
 	struct lttng_enabler *enabler = objd_private(objd);
 
 	if (enabler)
-		return lttng_ust_objd_unref(enabler->chan->objd);
+		return lttng_ust_objd_unref(enabler->chan->objd, 0);
 	return 0;
 }
 
